#define STRICT

// Includes standard Windows
#include <windows.h>
#include <windowsx.h>
#include <time.h>
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <stdio.h>

// Includes D3D
#define  D3D_OVERLOADS
#include <ddraw.h>
#include <d3d.h>
#include <d3dx.h>

// Includes utilitaires D3D
#include "d3dmath.h"
#include "d3dutil.h"
#include "D3DEnum.h"

// Ids Resources
#include "resource.h"

// Constantes
#include "const.h"

// Types
#include "types.h"

// Variables globales projet
#include "vars.h"

// Prototypes fonctions autres modules
#include "proto.h"

// Macros
#include "macros.h"

// Variables statiques au module
static BOOL bTracking = FALSE;							// Poursuite curseur en cours o/n (WM_LBUTTONDOWN / WM_MOUSEMOVE)
static WPARAM wModifier;								// Mmoris sur le LBUTTONDOWN : SHIFT / CTRL / ...
static HBRUSH hbrDkBlueSk, hbrLtBlueSk;						// Brosses de remplissage
static HPEN hpenWhite, hpenGrey, hpenCyan, hpenYellow, hpenPink, hpenDkBlueSk, hpenRed;	// Crayons
static HFONT hFont;


void vAddTextOutline(void)
{
#define XDC_HEADER	0
#define XDC_CURVE	1

	GLYPHMETRICS sGlyphMetrics;		// Rcuprer infos de kerning du caractre courant
	MAT2 sMatrix;					// Matrice identit (pourrait servir  faire tourner les caractres)
	WinThings *	psWinProps			// Pour rcuperer un HDC avec font
		= (WinThings *) GetWindowLong(hWndFace, 0);
	CHOOSEFONT cf;					// structure common dialog box 
	static LOGFONT lf;				// structure logical font 
	static DWORD rgbCurrent;		// current text color
	HFONT hfont, hfontPrev = NULL;
	float fFontScale = 1.f;

	// Initialize CHOOSEFONT
	ZeroMemory(&cf, sizeof(CHOOSEFONT));
	cf.lStructSize = sizeof (CHOOSEFONT);
	cf.hwndOwner = NULL;
	cf.lpLogFont = &lf;
	cf.rgbColors = rgbCurrent;
	cf.Flags = CF_BOTH | CF_TTONLY | CF_NOSCRIPTSEL | CF_INITTOLOGFONTSTRUCT;
	strcpy(lf.lfFaceName, "Arial");

	if (ChooseFont(&cf)==TRUE)
	{
		hfont = CreateFontIndirect(cf.lpLogFont);
		hfontPrev = (HFONT) SelectObject(psWinProps ->hdcDBuf, hfont);
		fFontScale = cf.iPointSize / 40.f;
	}
 
	// Saisir la chane
	if (DialogBoxParam(hInst, MAKEINTRESOURCE(IDD_STRING), NULL, bStringDlgProc, 0))
		return;

	// Rgler la matrice sur l'identit (pour ne pas faire tourner les caractres)
	sMatrix.eM11.value = sMatrix.eM22.value = 1;
	sMatrix.eM12.value = sMatrix.eM21.value =
	sMatrix.eM11.fract = sMatrix.eM22.fract = sMatrix.eM12.fract = sMatrix.eM21.fract = 0;

	// Pour chaque caractre de la chane
	for (char *pChar = cString ; *pChar ; pChar++)
	{
		TTPOLYGONHEADER	*pPolyHeader;
		TTPOLYCURVE *pPolyCurve;
		DWORD iBufSize;
		BOOL bState = XDC_HEADER;
		POINTFX *pPoints;
		BYTE *pFullBuffer, *pBlock, *pCurve;
		int iVert0, iVert1, iVert2;
		float fValue, fKmin = 100000.f, fKmax = -100000.f;
		D3DVECTOR dVec = Cursor1;
		
		// Suppression des accents
		if (strchr("", *pChar))	*pChar = 'a';
		if (strchr("", *pChar))	*pChar = 'e';
		if (strchr("", *pChar))	*pChar = 'i';
		if (strchr("", *pChar))	*pChar = 'o';
		if (strchr("", *pChar))	*pChar = 'u';
		if (strchr("", *pChar))	*pChar = 'y';

		// Si le caractre est un espace, on fait seulement avancer le curseur
		if (*pChar == ' ')
		{
			switch(lWActive)
			{
				case XDC_WID_TOP : // X et Z
				case XDC_WID_FACE : // X et Y
					Cursor1.x += fFontScale * (sGlyphMetrics.gmBlackBoxX / 2.f);
					break;
						
				case XDC_WID_SIDE : // Z et Y
					Cursor1.z += fFontScale * (sGlyphMetrics.gmBlackBoxX / 2.f);
					break;
			}
			continue;
		}

		// Rcuprer la taille du buffer de courbes dfinissant le caractre courant
		iBufSize = GetGlyphOutline(psWinProps ->hdcDBuf, *pChar, GGO_NATIVE, &sGlyphMetrics, 0, NULL, &sMatrix);
		
		// Si l'outliner n'est pas capable de gnrer le buffer pour le caractre, au suivant.
		if (iBufSize == 0xFFFFFFFF) continue;

		// Allouer le buffer de courbes
		if (!(pBlock = pFullBuffer = (unsigned char *) malloc(iBufSize)))
			continue;

		// Rcuprer le buffer de courbes
		GetGlyphOutline(psWinProps ->hdcDBuf, *pChar, GGO_NATIVE, &sGlyphMetrics, iBufSize, pFullBuffer, &sMatrix);
		
		// Pour chaque bloc en-tte / courbe(s)
		while (pBlock < pFullBuffer + iBufSize - 1)
		{
			switch(bState)
			{
				case XDC_HEADER: // Si on est sur un en-tte
					// Rcuprer un pointeur sur le header
					pPolyHeader = (TTPOLYGONHEADER *) pBlock;
					
					// Crer le point de dpart de l'enveloppe du caractre, et mmoriser son indice
					switch(lWActive)
					{
						case XDC_WID_TOP : // X et Z
							dVec.x = Cursor1.x + fFontScale * (pPolyHeader -> pfxStart.x.value + (float) (pPolyHeader -> pfxStart.x.fract) / 65536.f);
							dVec.z = Cursor1.z + fFontScale * (pPolyHeader -> pfxStart.y.value + (float) (pPolyHeader -> pfxStart.y.fract) / 65536.f);
							break;
								
						case XDC_WID_FACE : // X et Y
							dVec.x = Cursor1.x + fFontScale * (pPolyHeader -> pfxStart.x.value + (float) (pPolyHeader -> pfxStart.x.fract) / 65536.f);
							dVec.y = Cursor1.y + fFontScale * (pPolyHeader -> pfxStart.y.value + (float) (pPolyHeader -> pfxStart.y.fract) / 65536.f);
							break;
								
						case XDC_WID_SIDE : // Z et Y
							dVec.z = Cursor1.z + fFontScale * (pPolyHeader -> pfxStart.x.value + (float) (pPolyHeader -> pfxStart.x.fract) / 65536.f);
							dVec.y = Cursor1.y + fFontScale * (pPolyHeader -> pfxStart.y.value + (float) (pPolyHeader -> pfxStart.y.fract) / 65536.f);
							break;
					}
					iVert0 = iVert1 = iMakeVertex(dVec, XDC_ALLOWSAME);
					Vertices[iVert0].bSelected = TRUE;

					// Dcaler le pointeur de bloc sur le bloc courant (qui est une courbe)
					pBlock += sizeof(TTPOLYGONHEADER);
					bState = XDC_CURVE;

					// Special SG Kludge : ??? la taille indique pour le buffer est apparemment 4 x nb_headers plus grosse que la ralit...
					iBufSize -= 4;
					break;

				case XDC_CURVE: // Si on est sur une courbe
					// Rcuprer un pointeur sur la premire courbe du bloc
					pCurve = pBlock;

					// Dcaler le pointeur de bloc sur le bloc courant (qui est une courbe ou un header)
					pBlock += ((pPolyHeader -> cb) - sizeof(TTPOLYGONHEADER));
					
					// Pour chaque courbe du bloc
					do
					{
						// Rcuprer un pointeur sur la structure courbe
						pPolyCurve = (TTPOLYCURVE *) pCurve;

						// Calculer la taille de la courbe (pour ensuite passer  la suivante dansle bloc)
						int iSize = pPolyCurve -> cpfx * sizeof(pPolyCurve -> apfx) + sizeof(pPolyCurve -> cpfx) + sizeof(pPolyCurve -> wType);
						
						// Rcuprer un pointeur sur le tableau de points de la courbe
						pPoints = &(pPolyCurve -> apfx[0]);

						// Si la courbe est d'un type connu
						if (	pPolyCurve -> wType == TT_PRIM_LINE
							||	pPolyCurve -> wType == TT_PRIM_QSPLINE
							||	pPolyCurve -> wType == TT_PRIM_CSPLINE)
							// Pour chaque point de la courbe
							for (int iPoint = 0 ; iPoint < pPolyCurve -> cpfx ; iPoint++)
							{
								// Dessiner le point et tracer une arte avec le prcdent
								switch(lWActive)
								{
									case XDC_WID_TOP : // X et Z
										dVec.x = Cursor1.x + fFontScale * (pPoints[iPoint].x.value + (float) (pPoints[iPoint].x.fract) / 65536.f);
										dVec.z = Cursor1.z + fFontScale * (pPoints[iPoint].y.value + (float) (pPoints[iPoint].y.fract) / 65536.f);
										break;
											
									case XDC_WID_FACE : // X et Y
										dVec.x = Cursor1.x + fFontScale * (pPoints[iPoint].x.value + (float) (pPoints[iPoint].x.fract) / 65536.f);
										dVec.y = Cursor1.y + fFontScale * (pPoints[iPoint].y.value + (float) (pPoints[iPoint].y.fract) / 65536.f);
										break;
											
									case XDC_WID_SIDE : // Z et Y
										dVec.z = Cursor1.z + fFontScale * (pPoints[iPoint].x.value + (float) (pPoints[iPoint].x.fract) / 65536.f);
										dVec.y = Cursor1.y + fFontScale * (pPoints[iPoint].y.value + (float) (pPoints[iPoint].y.fract) / 65536.f);
										break;
								}
								iVert2 = iMakeVertex(dVec, XDC_ALLOWSAME);
								iMakeEdge(iVert2, iVert1);

								// Shifter le prcdent
								iVert1 = iVert2;

								// Ajuster les mini / maxi du kerning pour ce caractre
								fValue = fFontScale * (pPoints[iPoint].x.value + (float) (pPoints[iPoint].x.fract) / 65536.f);
								if (fKmin > fValue) fKmin = fValue;
								if (fKmax < fValue) fKmax = fValue;

								fValue = fFontScale * (pPolyHeader -> pfxStart.x.value + (float) (pPolyHeader -> pfxStart.x.fract) / 65536.f);
								if (fKmin > fValue) fKmin = fValue;
								if (fKmax < fValue) fKmax = fValue;
							}

						// Passer  la courbe suivante dans le bloc
						pCurve += iSize;

					} while (pCurve < pBlock && pCurve < pFullBuffer + iBufSize);

					// Refermer la courbe si elle ne l'est pas encore
					//if (pPolyCurve -> wType == TT_PRIM_LINE)
					if (-1 == iFindEdge(iVert2, iVert0))
						iMakeEdge(iVert2, iVert0);

					// Le prochain bloc est un header
					bState = XDC_HEADER;
					break;
			}
		}

		// Librer le buffer de courbes
		free(pFullBuffer);

		// Avancer le curseur
		switch(lWActive)
		{
			case XDC_WID_TOP : // X et Z
			case XDC_WID_FACE : // X et Y
				Cursor1.x += 1.15f * (fKmax - fKmin);
				break;
					
			case XDC_WID_SIDE : // Z et Y
				Cursor1.z += 1.15f * (fKmax - fKmin);
				break;
		}
	}

	// Remettre la fonte par dfaut du DC si on l'a change
	if (hfontPrev)
	{
		SelectObject(psWinProps ->hdcDBuf, hfontPrev);
		
		// Librer la fonte cre
		DeleteFont(hfont);
	}

}
