#define STRICT

// Includes standard Windows
#include <windows.h>
#include <windowsx.h>
#include <time.h>
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <stdio.h>

// Includes D3D
#define  D3D_OVERLOADS
#include <ddraw.h>
#include <d3d.h>

// Includes utilitaires D3D
#include "d3dmath.h"
#include "d3dutil.h"
#include "D3DEnum.h"

#include <d3dx.h>

// Ids Resources
#include "resource.h"

// Constantes
#include "const.h"

// Types
#include "types.h"

// Variables globales projet
#include "vars.h"

// Prototypes fonctions autres modules
#include "proto.h"

// Macros
#include "macros.h"

//************************************* C O M M U N S ***********************
void vDeleteObjects(void)
{
	// RAZ points
	for (int iVert = 0 ; iVert < XDC_NUMPT ; iVert++)
	{
		Vertices[iVert].bEnabled = FALSE;
		Vertices[iVert].bHidden = FALSE;
		Vertices[iVert].bSelected = FALSE;
		Vertices[iVert].bSmooth = FALSE;
		Vertices[iVert].iDepend = 0;
	}
	iVertFirstAvailable = 0;
	iVertLastUsed = -1;

	// RAZ artes
	for (int iEdge = 0 ; iEdge < XDC_NUMEDGE ; iEdge++)
	{
		Edges[iEdge].bEnabled = FALSE;
		Edges[iEdge].bHidden = FALSE;
	}
	iEdgeFirstAvailable = 0;
	iEdgeLastUsed = -1;

	// RAZ lampes
	for (int iLamp = 0 ; iLamp < XDC_NUMLAMP ; iLamp++)
		if (Lampes[iLamp].bEnabled) bDeleteLamp(iLamp);

	iLampFirstAvailable = 0;
	iLampLastUsed = -1;

	// RAZ facettes
	for (int iTriangle = 0 ; iTriangle < XDC_NUMTRI ; iTriangle++)
	{
		Triangles[iTriangle].bEnabled = FALSE;
		Triangles[iTriangle].bHidden = FALSE;
	}
	iTriaFirstAvailable = 0;
	iTriaLastUsed = -1;

	// RAZ Objets D3D
	vDeleteD3DObjects();
}

//************************************** V E R T E X *************************
int iFindVertex(D3DVECTOR p, int iPrevious)
{
	int iCnt;
	float fTolerance = (float) fabs(fXmax - fXmin) / XDC_ZONE;
	
	// Rechercher le point
	for (iCnt = max(iPrevious + 1, 0) ;
		    (iCnt <= iVertLastUsed)
		 && (!(   (Vertices[iCnt].bEnabled == TRUE)
			   && (Vertices[iCnt].bHidden == FALSE)
			   && (fabs(Vertices[iCnt].vPoint.x - p.x) < fTolerance)
			   && (fabs(Vertices[iCnt].vPoint.y - p.y) < fTolerance)
			   && (fabs(Vertices[iCnt].vPoint.z - p.z) < fTolerance)
			  ))
		 ;
		 iCnt++) ;

	// Si point trouv alors renvoyer son indice
	if (iCnt <= iVertLastUsed)
		return(iCnt);

	// Sinon renvoyer -1
	return -1;
}

int iMakeVertex(D3DVECTOR p, BOOL bSame)
{
	int iCnt;
	
	// Rechercher le premier index libre
	for (iCnt = bSame ? 0 : iVertFirstAvailable ;
		    (iCnt <= iVertLastUsed)
		 && (   ((!bSame) && (Vertices[iCnt].bEnabled == TRUE))
		     || ((bSame) && (!((Vertices[iCnt].bEnabled == TRUE) && (Vertices[iCnt].bHidden == FALSE) && (Magnitude(Vertices[iCnt].vPoint - p) < 10.f * g_EPSILON))))
			) ;
		 iCnt++) ;

	// Si limite tableau atteinte alors renvoyer erreur
	if (iCnt >= XDC_NUMPT)
	{
		vTrace("*** E0016 : Plus de slots vertex libres");
		return -1;
	}

	// Si point trouv dans tableau alors le renvoyer
	if (bSame)
	{
		if ((iCnt < XDC_NUMPT) && (Vertices[iCnt].bEnabled == TRUE) && (Vertices[iCnt].bHidden == FALSE) && (Magnitude(Vertices[iCnt].vPoint - p) < 10.f * g_EPSILON))
			return(iCnt);
		else
			return(iMakeVertex(p, XDC_FORCENEW));
	}

	// Crer un nouveau point
	Vertices[iCnt].vPoint = p;
	Vertices[iCnt].bSelected = FALSE;
	Vertices[iCnt].bHidden = FALSE;
	Vertices[iCnt].bEnabled  =TRUE;

	while (Vertices[iVertFirstAvailable].bEnabled == TRUE) iVertFirstAvailable++;
	if (iVertLastUsed < iCnt) iVertLastUsed = iCnt;

	return iCnt;
}

BOOL bVertInEdge(int iVert, int iEdge)
{
	if (    (Edges[iEdge].iSommets[0] == iVert)
		 || (Edges[iEdge].iSommets[1] == iVert)
	   )
		return TRUE;
	else
		return FALSE;
}

BOOL bVertInTriangle(int iVert, int iTriangle)
{
	if (    (Triangles[iTriangle].iSommets[0] == iVert)
		 || (Triangles[iTriangle].iSommets[1] == iVert)
		 || (Triangles[iTriangle].iSommets[2] == iVert)
	   )
		return TRUE;
	else
		return FALSE;
}

int iEdgesPerVert(int iVert)
{
	int iCnt, jCnt = 0;

	for (iCnt = 0 ; iCnt <= iVertLastUsed ; iCnt++)
		if (bVertInEdge(iVert, iCnt))
			jCnt++;

	return jCnt;
}

int iTrianglesPerVert(int iVert)
{
	int iCnt, jCnt = 0;

	for (iCnt = 0 ; iCnt <= iVertLastUsed ; iCnt++)
		if (bVertInTriangle(iVert, iCnt))
			jCnt++;

	return jCnt;
}

BOOL bDeleteVertex(int iVert)
{
	int iCnt;

	// Si le point n'est as enabled alors ne rien faire
	if (Vertices[iVert].bEnabled == FALSE) return FALSE;

	// Disabler le point
	Vertices[iVert].bEnabled = FALSE;
	
	// Mettre  jour les indices globaux du tableau des vertices
	if (iVert == iVertLastUsed) while (!Vertices[--iVertLastUsed].bEnabled) ;
	if (iVert < iVertFirstAvailable) iVertFirstAvailable = iVert;

	// Supprimer toutes les artes desquelles le sommet supprim est une extrmit
	for (iCnt = 0 ; iCnt <= iEdgeLastUsed ; iCnt++)
		if (bVertInEdge(iVert, iCnt))
			bDeleteEdge(iCnt);

	// Pour tout triangle duquel le sommet supprim est un sommet, transformer le triangle en une arte entre les 2 sommets restants
	for (iCnt = 0 ; iCnt <= iTriaLastUsed ; iCnt++)
		if (bVertInTriangle(iVert, iCnt))
		{
			/*
			if (Triangles[iCnt].iSommets[0] == iVert) iMakeEdge(Triangles[iCnt].iSommets[1], Triangles[iCnt].iSommets[2]);
			if (Triangles[iCnt].iSommets[1] == iVert) iMakeEdge(Triangles[iCnt].iSommets[0], Triangles[iCnt].iSommets[2]);
			if (Triangles[iCnt].iSommets[2] == iVert) iMakeEdge(Triangles[iCnt].iSommets[0], Triangles[iCnt].iSommets[1]);
			*/
			bDeleteTriangle(iCnt);
		}

	return TRUE;
}

BOOL bSmoothNorm(int iVertex)
{
	if (Vertices[iVertex].bEnabled == FALSE) return FALSE;
	if (Vertices[iVertex].bSmooth == FALSE) return FALSE;
	if (Vertices[iVertex].iDepend == 0) return FALSE;

	D3DVECTOR vn = D3DVECTOR(0.f, 0.f, 0.f);

	for (int iTriangle = 0 ; iTriangle <= iTriaLastUsed ; iTriangle++)
	{
		if (Triangles[iTriangle].bEnabled == FALSE) continue;
		if (!bVertInTriangle(iVertex, iTriangle)) continue;

		D3DVECTOR vc;
/*
		short iS0 = Triangles [iTriangle].iSommets[0],
			  iS1 = Triangles [iTriangle].iSommets[1],
			  iS2 = Triangles [iTriangle].iSommets[2];

		D3DVECTOR
				v01 = Vertices[iS1].vPoint - Vertices[iS0].vPoint,
				v02 = Vertices[iS2].vPoint - Vertices[iS0].vPoint,
				v12 = Vertices[iS2].vPoint - Vertices[iS1].vPoint,
				vn0, vn1, vn2;

		vn0 = Normalize(CrossProduct(v01, v02));
		vn1 = Normalize(CrossProduct(v01, v12));
		vn2 = Normalize(CrossProduct(v02, v12));

*/
		for (int iCnt = 0 ; iCnt < 3 ; iCnt++)
			if (iVertex == Triangles [iTriangle].iSommets[iCnt])
				vc = D3DVECTOR(
					Triangles [iTriangle].pSommets[iCnt].nx,
					Triangles [iTriangle].pSommets[iCnt].ny,
					Triangles [iTriangle].pSommets[iCnt].nz);

/*
		if (iVertex == iS0) vc = vn0;
		if (iVertex == iS1) vc = vn1;
		if (iVertex == iS2) vc = vn2;

		float fdp = DotProduct(vn, vc);

		if (fdp >= 0)
*/
			vn += vc;
/*
		else
			vn -= vc;
*/
	}

	Vertices[iVertex].vNormal = Normalize(vn);

	return TRUE;
}

BOOL bIsVertexSelected(int iVertex)
{
	return
		Vertices[iVertex].bEnabled && Vertices[iVertex].bSelected;
}

//************************************** L A M P E S *************************
int iFindLamp(D3DVECTOR p, int iPrevious)
{
	int iCnt;
	float fTolerance = (float) fabs(fXmax - fXmin) / XDC_ZONE;
	
	// Rechercher la lampe
	for (iCnt = max(iPrevious + 1, 0) ;
		    (iCnt <= iLampLastUsed)
		 && (!(   (Lampes[iCnt].bEnabled == TRUE)
			   && (fabs(Lampes[iCnt].lLamp.dvPosition.x - p.x) < fTolerance)
			   && (fabs(Lampes[iCnt].lLamp.dvPosition.y - p.y) < fTolerance)
			   && (fabs(Lampes[iCnt].lLamp.dvPosition.z - p.z) < fTolerance)
			  ))
		 ;
		 iCnt++) ;

	// Si point trouv alors renvoyer son indice
	if (iCnt <= iLampLastUsed)
		return(iCnt);

	// Sinon renvoyer -1
	return -1;
}

int iMakeLamp(
    D3DLIGHTTYPE    dltType,            /* Type of light source */
    D3DCOLORVALUE   dcvDiffuse,         /* Diffuse color of light */
    D3DCOLORVALUE   dcvSpecular,        /* Specular color of light */
    D3DCOLORVALUE   dcvAmbient,         /* Ambient color of light */
    D3DVECTOR       dvPosition,         /* Position in world space */
    D3DVECTOR       dvDirection,        /* Direction in world space */
    D3DVALUE        dvRange,            /* Cutoff range */
    D3DVALUE        dvFalloff,          /* Falloff */
    D3DVALUE        dvAttenuation0,     /* Constant attenuation */
    D3DVALUE        dvAttenuation1,     /* Linear attenuation */
    D3DVALUE        dvAttenuation2,     /* Quadratic attenuation */
    D3DVALUE        dvTheta,            /* Inner angle of spotlight cone */
    D3DVALUE        dvPhi               /* Outer angle of spotlight cone */
)
{
	int iCnt;
	
	// Rechercher le premier index libre
	for (iCnt = iLampFirstAvailable ;
		    (iCnt <= iLampLastUsed)
		 && (Lampes[iCnt].bEnabled == TRUE) ;
		 iCnt++) ;

	// Si limite tableau atteinte alors renvoyer erreur
	if (iCnt >= XDC_NUMLAMP)
	{
		vTrace("*** E0017 : Plus de slots vertex libres");
		return -1;
	}

	// Crer une nouvelle lampe
	ZeroMemory(&Lampes[iCnt].lLamp, sizeof(D3DLIGHT7));
	Lampes[iCnt].bEnabled = TRUE;
	Lampes[iCnt].bSelected = FALSE;
	Lampes[iCnt].bLit = TRUE;
    Lampes[iCnt].lLamp.dltType = dltType; // D3DLIGHT_POINT ou D3DLIGHT_DIRECTIONAL ou D3DLIGHT_SPOT
    Lampes[iCnt].lLamp.dcvDiffuse = dcvDiffuse;
    Lampes[iCnt].lLamp.dcvSpecular = dcvSpecular;
    Lampes[iCnt].lLamp.dcvAmbient = dcvAmbient;
	Lampes[iCnt].lLamp.dvPosition = dvPosition;
	Lampes[iCnt].lLamp.dvDirection = dvDirection;
    Lampes[iCnt].lLamp.dvRange = dvRange;
    Lampes[iCnt].lLamp.dvFalloff = dvFalloff;
    Lampes[iCnt].lLamp.dvAttenuation0 = dvAttenuation0; 
    Lampes[iCnt].lLamp.dvAttenuation1 = dvAttenuation1; 
    Lampes[iCnt].lLamp.dvAttenuation2 = dvAttenuation2; 
    Lampes[iCnt].lLamp.dvTheta = dvTheta; 
    Lampes[iCnt].lLamp.dvPhi = dvPhi; 

#ifndef NO3D
	lpd3dDevice->SetLight(iCnt, &Lampes[iCnt].lLamp);
	lpd3dDevice->LightEnable(iCnt, Lampes[iCnt].bLit);
#endif

	while (Lampes[iLampFirstAvailable].bEnabled == TRUE) iLampFirstAvailable++;
	if (iLampLastUsed < iCnt) iLampLastUsed = iCnt;

	return iCnt;
}

BOOL bUpdateLamp(int iLamp)
{
	// Si la lampe n'est pas enabled alors ne rien faire
	if (Lampes[iLamp].bEnabled == FALSE) return FALSE;

#ifndef NO3D
	lpd3dDevice->SetLight(iLamp, &Lampes[iLamp].lLamp);
	lpd3dDevice->LightEnable(iLamp, Lampes[iLamp].bLit);
#endif

	return TRUE;
}

BOOL bDeleteLamp(int iLamp)
{
	// Si la lampe n'est pas enabled alors ne rien faire
	if (Lampes[iLamp].bEnabled == FALSE) return FALSE;

	// Disabler la lampe
#ifndef NO3D
	lpd3dDevice->LightEnable(iLamp, FALSE);
#endif
	Lampes[iLamp].bEnabled = FALSE;
	Lampes[iLamp].bLit = FALSE;
	
	// Mettre  jour les indices globaux du tableau des lampes
	if (iLamp == iLampLastUsed) while (!Lampes[--iLampLastUsed].bEnabled) ;
	if (iLamp < iLampFirstAvailable) iLampFirstAvailable = iLamp;

	return TRUE;
}

//************************************** E D G E S ***************************
int iFindEdge(int i1, int i2)
{
	int iCnt;

	for (iCnt = 0 ; iCnt <= iEdgeLastUsed ; iCnt++)
		if (    ((Edges[iCnt].iSommets[0] == i1) || (Edges[iCnt].iSommets[1] == i1))
			 && ((Edges[iCnt].iSommets[0] == i2) || (Edges[iCnt].iSommets[1] == i2))
		   )
		   return iCnt;

	return -1;
}

int iMakeEdge(int i1, int i2)
{
	int iCnt;

	// Rechercher un slot d'arte libre
	for (iCnt = iEdgeFirstAvailable ; (iCnt <= iEdgeLastUsed) && (Edges [iCnt].bEnabled == TRUE) ; iCnt++);

	// Si pas trouv alors fail
	if (iCnt >= XDC_NUMEDGE)
	{
		vTrace("*** E0018 : Plus de slots artes libres");
		return -1;
	}

	// Crer l'arte
	Edges [iCnt].iSommets[0] = i1;
	Edges [iCnt].iSommets[1] = i2;

	Edges[iCnt].bEnabled = TRUE;
	Edges[iCnt].bHidden = FALSE;

	while (Edges[iEdgeFirstAvailable].bEnabled == TRUE) iEdgeFirstAvailable++;
	if (iEdgeLastUsed < iCnt) iEdgeLastUsed = iCnt;
	
	return iCnt;
}

BOOL bDeleteEdge(int iEdge)
{
	// Si l'arte n'est pas enabled alors ne rien faire
	if (Edges[iEdge].bEnabled == FALSE) return FALSE;

	// Disabler l'arte
	Edges[iEdge].bEnabled = FALSE;
	
	// Mettre  jour les indices globaux du tableau des artes
	if (iEdge == iEdgeLastUsed) while (!Edges[--iEdgeLastUsed].bEnabled) ;
	if (iEdge < iEdgeFirstAvailable) iEdgeFirstAvailable = iEdge;

	return TRUE;
}

BOOL	bIsEdgeSelected(int iEdge)
{
	return
		Edges[iEdge].bEnabled &&
		bIsVertexSelected(Edges[iEdge].iSommets[0]) &&
		bIsVertexSelected(Edges[iEdge].iSommets[1]);
}

BOOL	bIsEdgePartiallySelected(int iEdge)
{
	return 
		Edges[iEdge].bEnabled &&
		(bIsVertexSelected(Edges[iEdge].iSommets[0]) ||
		 bIsVertexSelected(Edges[iEdge].iSommets[1]));
}

//************************************** T R I A N G L E S *******************
int iFindTriangle(int i1, int i2, int i3)
{
	int iCnt;

	for (iCnt = 0 ; iCnt <= iTriaLastUsed ; iCnt++)
		if (    ((Triangles[iCnt].iSommets[0] == i1) || (Triangles[iCnt].iSommets[1] == i1) || (Triangles[iCnt].iSommets[2] == i1))
			 && ((Triangles[iCnt].iSommets[0] == i2) || (Triangles[iCnt].iSommets[1] == i2) || (Triangles[iCnt].iSommets[2] == i2))
			 && ((Triangles[iCnt].iSommets[0] == i3) || (Triangles[iCnt].iSommets[1] == i3) || (Triangles[iCnt].iSommets[2] == i3))
		   )
		   return iCnt;

	return -1;
}

int iMakeTriangle(int i1, int i2, int i3, unsigned char bRouge, unsigned char bVert, unsigned char bBleu)
{
	int iCnt;

	// Rechercher un slot de triangle libre
	for (iCnt = iTriaFirstAvailable ; (iCnt <= iTriaLastUsed) && (Triangles [iCnt].bEnabled == TRUE) ; iCnt++);

	// Si pas trouv alors fail
	if (iCnt >= XDC_NUMTRI)
	{
		vTrace("*** E0019 : Plus de slots triangles libres");
		return -1;
	}

	// Crer le triangle
	Triangles [iCnt].iSommets[0] = i1;
	Triangles [iCnt].iSommets[1] = i2;
	Triangles [iCnt].iSommets[2] = i3;

	// Incrmenter le compteur de triangles des sommets
	Vertices[Triangles [iCnt].iSommets[0]].iDepend++;
	Vertices[Triangles [iCnt].iSommets[1]].iDepend++;
	Vertices[Triangles [iCnt].iSommets[2]].iDepend++;

	// Affecter couleur et aspect
	D3DMATERIAL7 *pMtrl = &Triangles[iCnt].mtrl;
	ZeroMemory(pMtrl, sizeof(D3DMATERIAL7));
	pMtrl -> ambient.a = 0.5f;
	pMtrl -> ambient.r = (float) bRouge / 255.f / 6.f;
	pMtrl -> ambient.g = (float) bVert / 255.f / 6.f;
	pMtrl -> ambient.b = (float) bBleu / 255.f / 6.f;

	pMtrl -> diffuse.a = 0.5f;
	pMtrl -> diffuse.r = (float) bRouge / 255.f;
	pMtrl -> diffuse.g = (float) bVert / 255.f;
	pMtrl -> diffuse.b = (float) bBleu / 255.f;

	pMtrl -> specular.a = 0.5f;
	pMtrl -> specular.r = 1.f;
	pMtrl -> specular.g = 1.f;
	pMtrl -> specular.b = 1.f;
	pMtrl -> power = 1.f;

	Triangles[iCnt].bEnabled = TRUE;
	Triangles[iCnt].bHidden = FALSE;

	// Mettre  jour les coordonnes du triangle D3D correspondant
	bUpdateD3DTri(iCnt);

	while (Triangles[iTriaFirstAvailable].bEnabled == TRUE) iTriaFirstAvailable++;
	if (iTriaLastUsed < iCnt) iTriaLastUsed = iCnt;
	
	return iCnt;
}

BOOL bDeleteTriangle(int iTriangle)
{
	// Si le triangle n'est pas enabled alors ne rien faire
	if (Triangles[iTriangle].bEnabled == FALSE) return FALSE;

	// Disabler le triangle
	Triangles[iTriangle].bEnabled = FALSE;
	
	// Mettre  jour les indices globaux du tableau des triangles
	if (iTriangle == iTriaLastUsed) while (!Triangles[--iTriaLastUsed].bEnabled) ;
	if (iTriangle < iTriaFirstAvailable) iTriaFirstAvailable = iTriangle;

	// Dcrmenter le compteur de triangles des sommets
	Vertices[Triangles [iTriangle].iSommets[0]].iDepend--;
	Vertices[Triangles [iTriangle].iSommets[1]].iDepend--;
	Vertices[Triangles [iTriangle].iSommets[2]].iDepend--;

	return TRUE;
}

BOOL bUpdateD3DTri(int iCnt)
{
	if (Triangles[iCnt].bEnabled == FALSE) return FALSE;
	
	short iS0 = Triangles [iCnt].iSommets[0],
		  iS1 = Triangles [iCnt].iSommets[1],
		  iS2 = Triangles [iCnt].iSommets[2];

	D3DVECTOR
			v01 = Vertices[iS1].vPoint - Vertices[iS0].vPoint,
			v02 = Vertices[iS2].vPoint - Vertices[iS0].vPoint,
			v12 = Vertices[iS2].vPoint - Vertices[iS1].vPoint,
			vn0, vn1, vn2,
			v1, v2, vp2, v3, v4;
	
	D3DMATRIX p;

	// Calculer la normale au point 0 : v3
	v1 = Normalize(v01);
	v2 = Normalize(v02);
	v3 = Normalize(CrossProduct(v01, v02));
	
	// On fait un repre orthonormal avec v3 et v1 (qui sont orthonorms) et leur produit vectoriel
	vp2 = Normalize(CrossProduct(v3, v1));

	// On fait la matrice de changement de repre 
	D3DUtil_SetIdentityMatrix(p);
	p._11 = v1.x;
	p._21 = v1.y;
	p._31 = v1.z;
	p._12 = vp2.x;
	p._22 = vp2.y;
	p._32 = vp2.z;
	p._13 = v3.x;
	p._23 = v3.y;
	p._33 = v3.z;

	// on calcule l'image de z dans le nouveau repre : v4
	D3DMath_VectorMatrixMultiply(v4, D3DVECTOR(0.f, 0.f, 1.f), p);

	// On vrifie que v4 est dans le mme sens que z, ce qui nous dit si le triangle est CW ou CCW
	float fSign = DotProduct(D3DVECTOR(0.f, 0.f, 1.f), v4);

	// On calcule les vecteurs normaux
	if (Vertices[iS0].bSmooth) { bSmoothNorm(iS0); vn0 = Vertices[iS0].vNormal; }
	else vn0 = Normalize(CrossProduct(v01, v02));

	if (Vertices[iS1].bSmooth) { bSmoothNorm(iS1); vn1 = Vertices[iS1].vNormal; }
	else vn1 = Normalize(CrossProduct(v01, v12));

	if (Vertices[iS2].bSmooth) { bSmoothNorm(iS2); vn2 = Vertices[iS2].vNormal; }
	else vn2 = Normalize(CrossProduct(v02, v12));

	// Crer les 2 triangles D3D (CW, CCW) en ordonnant les sommets pour avoir les CCW en 1er
	if (fSign >= 0)
	{
		Triangles [iCnt].pSommets[0] = D3DVERTEX(Vertices[iS0].vPoint, vn0, 0, 0);
		Triangles [iCnt].pSommets[1] = D3DVERTEX(Vertices[iS1].vPoint, vn1, 0, 0);
		Triangles [iCnt].pSommets[2] = D3DVERTEX(Vertices[iS2].vPoint, vn2, 0, 0);
		Triangles [iCnt].pSommets[3] = D3DVERTEX(Vertices[iS0].vPoint, -vn0, 0, 0);
		Triangles [iCnt].pSommets[4] = D3DVERTEX(Vertices[iS2].vPoint, -vn2, 0, 0);
		Triangles [iCnt].pSommets[5] = D3DVERTEX(Vertices[iS1].vPoint, -vn1, 0, 0);
	}
	else
	{
		Triangles [iCnt].pSommets[0] = D3DVERTEX(Vertices[iS0].vPoint, -vn0, 0, 0);
		Triangles [iCnt].pSommets[1] = D3DVERTEX(Vertices[iS2].vPoint, -vn2, 0, 0);
		Triangles [iCnt].pSommets[2] = D3DVERTEX(Vertices[iS1].vPoint, -vn1, 0, 0);
		Triangles [iCnt].pSommets[3] = D3DVERTEX(Vertices[iS0].vPoint, vn0, 0, 0);
		Triangles [iCnt].pSommets[4] = D3DVERTEX(Vertices[iS1].vPoint, vn1, 0, 0);
		Triangles [iCnt].pSommets[5] = D3DVERTEX(Vertices[iS2].vPoint, vn2, 0, 0);
	}

	return TRUE;
}

BOOL bSubdivideTriangle(int iTriangle)
{
	int iNewV[3], iNewT[4];
	
	if (Triangles[iTriangle].bEnabled == FALSE) return(FALSE);

	unsigned char uR = (unsigned char) (Triangles[iTriangle].mtrl.diffuse.r * 255.f),
				  uG = (unsigned char) (Triangles[iTriangle].mtrl.diffuse.g * 255.f),
				  uB = (unsigned char) (Triangles[iTriangle].mtrl.diffuse.b * 255.f);

	if (-1 == (iNewV[0] = iMakeVertex(
				(Vertices[Triangles[iTriangle].iSommets[0]].vPoint +
				 Vertices[Triangles[iTriangle].iSommets[1]].vPoint
				) / 2.f, XDC_ALLOWSAME))) return FALSE;

	if (-1 == (iNewV[1] = iMakeVertex(
				(Vertices[Triangles[iTriangle].iSommets[1]].vPoint +
				 Vertices[Triangles[iTriangle].iSommets[2]].vPoint
				) / 2.f, XDC_ALLOWSAME))) return FALSE;

	if (-1 == (iNewV[2] = iMakeVertex(
				(Vertices[Triangles[iTriangle].iSommets[2]].vPoint +
				 Vertices[Triangles[iTriangle].iSommets[0]].vPoint
				) / 2.f, XDC_ALLOWSAME))) return FALSE;

    if (-1 == (iNewT[0] = iMakeTriangle(iNewV[0], iNewV[1], iNewV[2], uR, uG, uB))) return FALSE;
    if (-1 == (iNewT[1] = iMakeTriangle(iNewV[0], iNewV[1], Triangles[iTriangle].iSommets[1], uR, uG, uB))) return FALSE;
    if (-1 == (iNewT[2] = iMakeTriangle(iNewV[1], iNewV[2], Triangles[iTriangle].iSommets[2], uR, uG, uB))) return FALSE;
    if (-1 == (iNewT[3] = iMakeTriangle(iNewV[0], iNewV[2], Triangles[iTriangle].iSommets[0], uR, uG, uB))) return FALSE;

	Vertices[iNewV[0]].bSelected =
	Vertices[iNewV[1]].bSelected =
	Vertices[iNewV[2]].bSelected = Vertices[Triangles[iTriangle].iSommets[0]].bSelected;

	return bDeleteTriangle(iTriangle);
}

BOOL	bIsTriangleSelected(int iTriangle)
{
	return
		Triangles[iTriangle].bEnabled &&
		bIsVertexSelected(Triangles[iTriangle].iSommets[0]) &&
		bIsVertexSelected(Triangles[iTriangle].iSommets[1]) &&
		bIsVertexSelected(Triangles[iTriangle].iSommets[2]);
}

BOOL	bIsTrianglePartiallySelected(int iTriangle)
{
	return 
		Triangles[iTriangle].bEnabled &&
		(bIsVertexSelected(Triangles[iTriangle].iSommets[0]) ||
		 bIsVertexSelected(Triangles[iTriangle].iSommets[1]) ||
		 bIsVertexSelected(Triangles[iTriangle].iSommets[2]));
}

//********************************************** O U T I L S *************************
// Propager de faon **non** rcursive la d/slection d'un vertex aux vertices connexes ( travers edges ou triangles)
void vPropagateSelection(void)
{
	register BOOL bReste;

	bReste = TRUE;
	while (bReste)
	{
		bReste = FALSE;

		for (register int iEdge = 0 ; iEdge <= iEdgeLastUsed ; iEdge++)
			if (bIsEdgePartiallySelected(iEdge) && !bIsEdgeSelected(iEdge))
				bReste =
				Vertices[Edges[iEdge].iSommets[0]].bSelected = 
				Vertices[Edges[iEdge].iSommets[1]].bSelected = TRUE;
	}

	bReste = TRUE;
	while (bReste)
	{
		bReste = FALSE;

		for (register int iTria = 0 ; iTria <= iTriaLastUsed ; iTria++)
			if (bIsTrianglePartiallySelected(iTria) && !bIsTriangleSelected(iTria))
				bReste =
				Vertices[Triangles[iTria].iSommets[0]].bSelected = 
				Vertices[Triangles[iTria].iSommets[1]].bSelected = 
				Vertices[Triangles[iTria].iSommets[2]].bSelected = TRUE;
	}
}

// Cloner la slection
BOOL bCloneSelection(D3DVECTOR vShift, BOOL bCloneMode)
{
	int iCnt, jCnt, iNumSelected, *iIndex;

	// Compter les points slectionns
	iNumSelected = 0;
	for (iCnt = 0 ; iCnt <= iVertLastUsed ; iCnt++)
		if (bIsVertexSelected(iCnt)) iNumSelected++;

	if (!iNumSelected) return FALSE;

	// Allouer la table de correspondance ancien / nouvel index
	iIndex = (int *) malloc(iNumSelected * 2 * sizeof(int));

	// Cloner les points slectionns en conservant la table de correspondance ancien / nouveau
	// et en dcalant les nouveaux de vShift
	jCnt = 0;
	for (iCnt = 0 ; iCnt <= iVertLastUsed ; iCnt++)
	{
		D3DVECTOR vVect = Vertices[iCnt].vPoint + vShift;

		if (!bIsVertexSelected(iCnt)) continue;

		// Remplir la table de correspondance ancien / nouveau
		iIndex[jCnt] = iCnt;
		iIndex[jCnt + 1] = iMakeVertex(vVect, XDC_FORCENEW);
		jCnt += 2;
	}

	// Cloner les artes ventuelles entre les anciens points
	for (iCnt = 0 ; iCnt <= iEdgeLastUsed ; iCnt++)
	{
		if (Edges[iCnt].bEnabled == FALSE) continue;

		int iOVert1 = Edges[iCnt].iSommets[0],
			iOVert2 = Edges[iCnt].iSommets[1],
			iCVert1 = -1,
			iCVert2 = -1;

		// Vrifier que les deux sommets originaux de l'arte sont slectionns
		if (bIsVertexSelected(iOVert1) && bIsVertexSelected(iOVert2))
		{
			// Alors il faut crer une nouvelle arte avec les sommets clons
			for (jCnt = 0 ; jCnt < iNumSelected ; jCnt++)
			{
				if (iIndex[jCnt * 2] == iOVert1) iCVert1 = iIndex[jCnt * 2 + 1];
				if (iIndex[jCnt * 2] == iOVert2) iCVert2 = iIndex[jCnt * 2 + 1];
			}
			if ((iCVert1 != -1) && (iCVert2 != -1))
			{
				iMakeEdge(iCVert1, iCVert2);

				// Si le mode de clonage est avec extrusion, crer les triangles de l'lvation
				if (bCloneMode == XDC_MODE_EXTRUDE)
				{
					iMakeTriangle(iOVert1, iCVert1, iCVert2, 0xFF, 0xFF, 0xFF);
					iMakeTriangle(iOVert1, iOVert2, iCVert2, 0xFF, 0xFF, 0xFF);
				}
			}
		}
	}

	// Cloner les triangles ventuels entre les diffrents points
	for (iCnt = 0 ; iCnt <= iTriaLastUsed ; iCnt++)
	{
		if (Triangles[iCnt].bEnabled == FALSE) continue;

		int iOVert1 = Triangles[iCnt].iSommets[0],
			iOVert2 = Triangles[iCnt].iSommets[1],
			iOVert3 = Triangles[iCnt].iSommets[2],
			iCVert1 = -1,
			iCVert2 = -1,
			iCVert3 = -1;

		// Vrifier que les trois sommets originaux du triangle sont slectionns
		if (bIsVertexSelected(iOVert1) && bIsVertexSelected(iOVert2) && bIsVertexSelected(iOVert3))
		{
			// Alors il faut crer un nouveau triangle avec les sommets clons
			for (jCnt = 0 ; jCnt < iNumSelected ; jCnt++)
			{
				if (iIndex[jCnt * 2] == iOVert1) iCVert1 = iIndex[jCnt * 2 + 1];
				if (iIndex[jCnt * 2] == iOVert2) iCVert2 = iIndex[jCnt * 2 + 1];
				if (iIndex[jCnt * 2] == iOVert3) iCVert3 = iIndex[jCnt * 2 + 1];
			}
			if ((iCVert1 != -1) && (iCVert2 != -1) && (iCVert3 != -1))
			{
				unsigned char uR = (unsigned char) (Triangles[iCnt].mtrl.diffuse.r * 255.f),
							  uG = (unsigned char) (Triangles[iCnt].mtrl.diffuse.g * 255.f),
							  uB = (unsigned char) (Triangles[iCnt].mtrl.diffuse.b * 255.f);

				iMakeTriangle(iCVert1, iCVert2, iCVert3, uR, uG, uB);

				// Si le mode de clonage est avec extrusion, crer les triangles de l'lvation
				if (bCloneMode == XDC_MODE_EXTRUDE)
				{
					iMakeTriangle(iOVert1, iCVert1, iCVert2, uR, uG, uB);
					iMakeTriangle(iOVert1, iOVert2, iCVert2, uR, uG, uB);
					iMakeTriangle(iOVert2, iCVert2, iCVert3, uR, uG, uB);
					iMakeTriangle(iOVert2, iOVert3, iCVert3, uR, uG, uB);
					iMakeTriangle(iOVert1, iCVert1, iCVert3, uR, uG, uB);
					iMakeTriangle(iOVert1, iOVert3, iCVert3, uR, uG, uB);
				}
			}
		}
	}

	// Inverser la slection (slectionner les nouveaux points)
	for (iCnt = 0 ; iCnt < iNumSelected ; iCnt++)
	{
		Vertices[iIndex[iCnt * 2    ]].bSelected = FALSE;
		Vertices[iIndex[iCnt * 2 + 1]].bSelected = TRUE;
	}

	// Librer la table de correspondance
	free(iIndex);

	return TRUE;
}

// Cacher les lments de la slection
void vHideSelection(void)
{
	int iVert, iCnt;

	for (iVert = 0 ; iVert <= iVertLastUsed ; iVert++)
	{
		if (bIsVertexSelected(iVert))
		{
			Vertices[iVert].bHidden = TRUE;
			Vertices[iVert].bSelected = FALSE;

			for (iCnt = 0 ; iCnt <= iEdgeLastUsed ; iCnt++)
			{
				if (Edges[iCnt].bEnabled == FALSE) continue;

				if (bVertInEdge(iVert, iCnt))
					Edges[iCnt].bHidden = TRUE;
			}
			for (iCnt = 0 ; iCnt <= iTriaLastUsed ; iCnt++)
			{
				if (Triangles[iCnt].bEnabled == FALSE) continue;

				if (bVertInTriangle(iVert, iCnt))
					Triangles[iCnt].bHidden = TRUE;
			}
		}
	}
}

// Rendre visibles les lments cachs
void vRevealHidden(void)
{
	int iCnt;

	for (iCnt = 0 ; iCnt <= iVertLastUsed ; iCnt++)
	{
		if (Vertices[iCnt].bEnabled == FALSE) continue;

		Vertices[iCnt].bHidden = FALSE;
	}

	for (iCnt = 0 ; iCnt <= iEdgeLastUsed ; iCnt++)
	{
		if (Edges[iCnt].bEnabled == FALSE) continue;

		Edges[iCnt].bHidden = FALSE;
	}

	for (iCnt = 0 ; iCnt <= iTriaLastUsed ; iCnt++)
	{
		if (Triangles[iCnt].bEnabled == FALSE) continue;

		Triangles[iCnt].bHidden = FALSE;
	}
}

// Dterminer le barycentre de la slection
D3DVECTOR vCenter(void)
{
	D3DVECTOR dVect = D3DVECTOR(0.f, 0.f, 0.f);
	int jCnt = 0;

	for (int iVert = 0 ; iVert <= iVertLastUsed ; iVert++)
		if (bIsVertexSelected(iVert))
		{
			// Mmoriser les coordonnes avant modification
			Vertices[iVert].vPointBack = Vertices[iVert].vPoint;
			
			// Compter les points
			jCnt++;

			// Additionner les coordonnes du point  celles du barycentre
			dVect.x += Vertices[iVert].vPoint.x;
			dVect.y += Vertices[iVert].vPoint.y;
			dVect.z += Vertices[iVert].vPoint.z;
		}

	if (!jCnt) 
	{
		vTrace("*** E0022 : Pas de slection");
		return dVect;
	}

	// Moyenner la somme des points
	dVect.x /= jCnt;
	dVect.y /= jCnt;
	dVect.z /= jCnt;

	return dVect;
}

// Garbage collektor !-)
void vCollect(void)
{
	int iShift, iVert, *iMapTo;

	// Rindexer les artes
	iShift = 0;

	for (int iEdge = 0 ; iEdge <= iEdgeLastUsed ; iEdge++)
	{
		if (Edges[iEdge].bEnabled == FALSE)
		{
			while (Edges[iEdge + iShift].bEnabled == FALSE)
			{
				iShift++;
				iEdgeLastUsed--;
			}
		}
		if (iShift)
		{
			Edges[iEdge] = Edges[iEdge + iShift];
			Edges[iEdge + iShift].bEnabled = FALSE;
		}
	}

	iEdgeFirstAvailable = iEdgeLastUsed + 1;

	// Rindexer les triangles
	iShift = 0;

	for (int iTria = 0 ; iTria <= iTriaLastUsed ; iTria++)
	{
		if (Triangles[iTria].bEnabled == FALSE)
		{
			while (Triangles[iTria + iShift].bEnabled == FALSE)
			{
				iShift++;
				iTriaLastUsed--;
			}
		}
		if (iShift)
		{
			Triangles[iTria] = Triangles[iTria + iShift];
			Triangles[iTria + iShift].bEnabled = FALSE;
		}
	}

	iTriaFirstAvailable = iTriaLastUsed + 1;

	// Compter les sommets inutiliss pour ne faire la suite que s'il y en a
	iShift = 0;

	for (iVert = 0 ; iVert <= iVertLastUsed ; iVert++)
		if (Vertices[iVert].bEnabled == FALSE)
			iShift++;

	if (!iShift)
		return;

	// Allouer la table de mapping
	if (!(iMapTo = (int *) malloc((iVertLastUsed + 1) * sizeof(int))))
	{
		vTrace("*** E0032 : erreur malloc table mapping GC");
		return;
	}

	// Prinitialiser la table de mapping
	for (iShift = 0 ; iShift <= iVertLastUsed ; iShift++)
		iMapTo[iShift] = iShift;

	// Rindexer les sommets
	iShift = 0;

	for (iVert = 0 ; iVert <= iVertLastUsed ; iVert++)
	{
		if (Vertices[iVert].bEnabled == FALSE)
		{
			while (Vertices[iVert + iShift].bEnabled == FALSE)
			{
				iShift++;
				iVertLastUsed--;
			}
		}

		if (iShift)
		{

			Vertices[iVert] = Vertices[iVert + iShift];
			Vertices[iVert + iShift].bEnabled = Vertices[iVert + iShift].bSelected = FALSE;
			iMapTo[iVert + iShift] = iVert;
		}

	}

	iVertFirstAvailable = iVertLastUsed + 1;

	// Rindexer les sommets dans les artes
	for(iShift = 0 ; iShift <= iEdgeLastUsed ; iShift++)
	{
		Edges[iShift].iSommets[0] = iMapTo[Edges[iShift].iSommets[0]];
		Edges[iShift].iSommets[1] = iMapTo[Edges[iShift].iSommets[1]];
	}

	// Rindexer les sommets dans les triangles
	for(iShift = 0 ; iShift <= iTriaLastUsed ; iShift++)
	{
		Triangles[iShift].iSommets[0] = iMapTo[Triangles[iShift].iSommets[0]];
		Triangles[iShift].iSommets[1] = iMapTo[Triangles[iShift].iSommets[1]];
		Triangles[iShift].iSommets[2] = iMapTo[Triangles[iShift].iSommets[2]];
	}

	// Librer la table
	free(iMapTo);
}
